#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <dlfcn.h>
#include <misc.h>
#pragma implementation
#include "tlmpweb.h"
#include "tlmpweb.m"
#include <tlmplib.h>
#include <translat.h>

static SSTRINGS tbvar,tbval;

static bool header_done = false;
static bool output_state = true;
static bool title_done = false;
static bool body_done=false;

extern "C" int __libc_write (int fd, const void *buf, unsigned len);

extern "C" int write (int fd, const void *buf, unsigned len)
{
	if (fd == 1 && output_state && !header_done){
		char head[200];
		int lenhead = snprintf (head,sizeof(head)-1
			,"X-Powered-By:  TLMP-web %s.%d\r\n"
			 "Content-Type: text/html\r\n"
			 "\r\n"
			,TLMPWEB_VERSION,TLMPWEB_RELEASE);
		__libc_write (1,head,lenhead);
		header_done = true;
	}
	return __libc_write (fd,buf,len);
}

/*
	Turn the output on an off.
	Return the previous state.
*/
bool tlmpweb_ctrloutput (bool on)
{
	static int null_fd = -1;
	static int real_fd = -1;
	if (null_fd == -1){
		null_fd = open ("/dev/null",O_WRONLY);
	}
	if (real_fd == -1){
		real_fd = dup (1);
	}
	bool ret = output_state;
	if (on){
		if (!output_state){
			fflush (stdout);
			dup2 (real_fd,1);
		}
	}else{
		if (output_state){
			fflush (stdout);
			dup2 (null_fd,1);
		}
	}
	output_state = on;
	return ret;
}


/*
	Add a line to the HTTP header
*/
void tlmpweb_header(const char *keyword, const char *value, ...)
{
	if (header_done) printf ("<br>tlmpweb_header() called too late.<br>\n");
	va_list list;
	va_start (list,value);
	printf ("%s: ",keyword);
	vprintf (value,list);
	va_end (list);
	printf ("\r\n");
}

/*
	Send the document type as part of the http header.
	This is the last entry sent in the HTTP header. Using tlmpweb_header
	will fail after this function is executed.
*/
void tlmpweb_doctype(const char *type)
{
	static bool done = false;
	if (!done){
		done = true;
		tlmpweb_header ("X-Powered-By","TLMP-web %s.%d"
			,TLMPWEB_VERSION,TLMPWEB_RELEASE);
		tlmpweb_header ("Content-Type","%s",type);
		printf ("\r\n");
		header_done = true;
	}
}

void tlmpweb_title(const char *title, ...)
{
	if (output_state && !title_done){
		title_done = true;
		tlmpweb_doctype ("text/html");
		printf ("<HTML>\n");
		printf ("<HEAD>\n");
		va_list list;
		va_start (list,title);
		printf ("<title>");
		vprintf (title,list);
		printf ("</title>");
		printf ("</head>\n");
	}
}

void tlmpweb_body (const char *bgcolor, const char *vlink)
{
	tlmpweb_body (bgcolor,vlink,"");
}

void tlmpweb_body (const char *bgcolor, const char *vlink, const char *opt)
{
	if (output_state && !body_done){
		printf ("<body");
		if (bgcolor != NULL) printf (" bgcolor=%s",bgcolor);
		if (vlink   != NULL) printf (" vlink=%s",vlink);
		printf (" %s>\n",opt);
		body_done = true;
	}
}

void tlmpweb_error (const char *msg, ...)
{
	if (!title_done) tlmpweb_title ("Some errors");
	va_list list;
	va_start (list,msg);
	printf ("tlmpweb: ");
	vprintf (msg,list);
	va_end (list);
}

static const char *curdir;
/*
	Load a shared object either in the lib sub-directory of the
	current html directory or in the /usr/lib/tlmp/lib directory

	Exit if the lib can't be found.
*/
void tlmpweb_loadlib (const char *name)
{
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"%s/%s.so",curdir,name);
	void *handle = dlopen (path,RTLD_LAZY|RTLD_GLOBAL);
	if (handle == NULL){
		snprintf (path,sizeof(path)-1,"/usr/lib/tlmp/lib/%s.so",name);
		void *handle = dlopen (path,RTLD_LAZY|RTLD_GLOBAL);
		if (handle == NULL){
			tlmpweb_error (MSG_U(E_NOLIB,"Can't find library %s.so (%s)\n")
				,name,dlerror());
			exit (-1);
		}else{
			void (*fct)() = (void (*)()) dlsym (handle,"tlmp_initmod");
			if (fct != NULL) (*fct)();
		}
	}
}

static void tlmpweb_loadlibs (const char *path)
{
	FILE *fin = fopen (path,"r");
	if (fin != NULL){
		char line[PATH_MAX];
		while (fgets(line,sizeof(line)-1,fin)!=NULL){
			strip_end (line);
			if (line[0] != '\0') tlmpweb_loadlib (line);
		}
		fclose (fin);
	}
}

/*
	Load all library enumerated in the .tlmplibs file of the current
	directory
*/
void tlmpweb_loadlibs ()
{
	char path[PATH_MAX];
	snprintf (path,sizeof(path)-1,"%s/.tlmplibs",curdir);
	if (file_exist (path)){
		tlmpweb_loadlibs (path);
	}else{
		const char *root = getenv ("DOCUMENT_ROOT");
		if (root != NULL){
			snprintf (path,sizeof(path)-1,"%s/.tlmplibs",root);
			if (file_exist(path)){
				tlmpweb_loadlibs (path);
			}
		}
	}
}
/*
	Copy the directory part of a path into "dir"
*/
static void tlmpweb_setdir (const char *path, char dir[PATH_MAX])
{
	strncpy (dir,path,PATH_MAX-1);
	dir[PATH_MAX-1] = '\0';
	char *pt = strrchr (dir,'/');
	if (pt != NULL) *pt = '\0';
}


static void tlmpweb_error0 (const char *msg)
{
	if (!title_done) tlmpweb_title ("Some errors");
	printf ("tlmpweb: %s",msg);
}

/*
	Return one web variable as a string
	is_def will be true if the variable was found.
*/
const char *t_varstr (const char *var, bool &is_def)
{
	const char *ret = "";
	int no = tbvar.lookup(var);
	is_def = false;
	if (no != -1){
		ret = tbval.getitem(no)->get();
		is_def = true;
	}
	return ret;
}
/*
	Return one web variable as a string
*/
const char *t_varstr (const char *var)
{
	bool is_def;
	return t_varstr (var,is_def);
}

/*
	Return true if a variable is defined
*/
bool t_var_isdef (const char *var)
{
	return tbvar.lookup(var) != -1;
}

/*
	Return one web variable as an int
*/
int t_varint (const char *var)
{
	return atoi(t_varstr(var));
}
/*
	Forget the value of a variable
*/
void t_var_setempty (const char *name)
{
	int no = tbvar.lookup(name);
	if (no != -1){
		SSTRING *s = tbvar.getitem(no);
		s->setempty();
	}
}


/*
	Return the name of the current web page
*/
const char *tlmpweb_curpage()
{
	const char *ret = "unknown";
	static const char *p = NULL;
	if (p == NULL){
		p = getenv ("PATH_INFO");
		if (p != NULL){
			const char *pt = strrchr(p,'/');
			if (pt != NULL) p = pt+1;
		}
	}
	if (p != NULL) ret = p;
	return ret;
}
/*
	Return the name of the host as seen by the client
*/
const char *tlmpweb_curhost ()
{
	const char *ret = getenv ("HTTP_HOST");
	if (ret == NULL) ret = "unknown";
	return ret;
}


/*
	Return the name of the current directory holding the current page
*/
const char *tlmpweb_curdir()
{
	static const char *ret = NULL;
	if (ret == NULL){
		const char *p = getenv ("PATH_INFO");
		if (p != NULL){
			char line[strlen(p)+1];
			strcpy (line,p);
			char *pt = strrchr(line,'/');
			if (pt != NULL) *pt = '\0';
			ret = strdup(line);
		}else{
			ret = "unknown";
		}
	}
	return ret;
}


static char hextoi (char asc)
{
	return isdigit(asc) ? asc - '0' : (toupper(asc) - 'A') + 10;
}

/*
	Extract and decode one line
*/
static const char *http_decode (const char *str, char *buf, int sizbuf)
{
	char *pt = buf;
	char *endbuf = buf + sizbuf-1;
	while (*str != '\0' && *str != '\n' && pt < endbuf){
		if (*str == '%'){
			str++;
			*pt++ = hextoi(*str++) * 16 + hextoi(*str++);
		}else if (*str == '+'){
			*pt++ = ' ';
			str++;
		}else{
			*pt++ = *str++;
		}
	}
	*pt = '\0';
	strip_end (buf);
	if (*str == '\n') str++;
	return str;
}

static void http_parsevar (const char *buf)
{
	while (1){
		char *pt = strchr(buf,'=');
		if (pt == NULL){
			break;
		}else{
			*pt++ = '\0';
			char *end = strchr(pt,'&');
			if (end != NULL){
				*end++ = '\0';
			}
			char var[200];
			char val[2000];
			http_decode (buf,var,sizeof(var));
			http_decode (pt,val,sizeof(val));
			tbvar.add (new SSTRING(var));
			tbval.add (new SSTRING(val));
			if (end == NULL) break;
			buf = end;
		}
	}
}

static void http_parsevar_env ()
{
	const char *buf = getenv ("QUERY_STRING");
	if (buf != NULL) http_parsevar (buf);
}


static void http_pickname (
	const char *line,
	const char *varname,
	SSTRING &val)
{
	const char *name = strstr(line,varname);
	if (name != NULL){
		name += strlen(varname);
		if (name[0] == '"') name++;
		const char *pt = name;
		while (*pt > ' ' && *pt != ';' && *pt != '"') pt++;
		int len = (int)(pt-name);
		val.setfrom (name,len);
	}
}

/*
	Get the temp file name associated with a field of type "file".
	Return an empty string if there is no file available.
*/
const char *tlmpweb_getfilename(const char *var)
{
	char tmp[strlen(var)+10];
	strcpy (tmp,"filename:");
	strcat (tmp,var);
	return t_varstr(tmp);
}

static SSTRINGS tbfile;		// List of temp file to delete at close time

class BUFINPUT{
	char buf[10000];
	int nb;
public:
	BUFINPUT();
	char *fgets(char *line, int size);
	int read (char *buf, int size);
	void unread (char *buf, int size);
};

BUFINPUT::BUFINPUT()
{
	nb = 0;
}

char *BUFINPUT::fgets (char *line, int size)
{
	char *ret = NULL;
	if (nb > 0){
		char *pt = strchr(buf,'\n');
		if (pt == NULL){
			// Not a complete line in the buffer, we need more
			strcpy (line,buf);
			fgets (line+nb,size-nb);
			nb = 0;
		}else{
			int len = (int)(pt-buf)+1;
			memcpy (line,buf,len);
			line[len] = '\0';
			nb -= len;
			if (nb > 0)	memcpy (buf,buf+len,nb);
		}
		ret = line;
	}else{
		ret = ::fgets (line,size,stdin);
	}
	return ret;
}

int BUFINPUT::read (char *line, int size)
{
	int ret = 0;
	if (nb > 0){
		if (size > nb) size = nb;
		memcpy (line,buf,size);
		if (nb > size){
			nb -= size;
			memcpy (buf,buf+size,nb);
		}else{
			nb = 0;
		}
		ret = size;
	}else{
		ret = fread (line,1,size,stdin);
	}
	return ret;
}

void BUFINPUT::unread (char *line, int len)
{
	memcpy (buf,line,len);
	nb = len;
	buf[nb] = '\0';
}	

static void http_parsevar_post()
{
	FILE *fout = fopen ("/tmp/toto.log","a");
	const char *content = getenv ("CONTENT_TYPE");
	const char *multi = "multipart/form-data;";
	// fprintf (fout,"content :%s:\n",content);
	// fflush (fout);
	if (content != NULL && strncasecmp(content,multi,strlen(multi))==0){
		const char *boundary=strstr(content,"boundary=");
		if (boundary != NULL){
			boundary += 9;
			int len_boundary = strlen(boundary);
			// fprintf (fout,"boundary :%s:\n",boundary);
			bool in_header = false;
			SSTRING variable,accum;
			char line[10000];
			FILE *fdata = NULL;
			BUFINPUT inp;
			while (inp.fgets(line,sizeof(line)-1)!=NULL){
				fprintf (fout,"POST[%d]: %s",in_header,line);
				if (line[0] == '-'
					&& line[1] == '-'
					&& strncmp(line+2,boundary,len_boundary)==0){
					in_header = true;
					if (variable.is_filled()){
						tbvar.add (new SSTRING(variable));
						tbval.add (new SSTRING(accum));
					}
					if (fdata != NULL){
						fclose (fdata);
						fdata = NULL;
					}
					variable.setfrom ("");
					accum.setfrom ("");
				}else if (in_header){
					// fprintf (fout,"in-header :%s",line);
					if (line[0] == '\n' || line[0] == '\r'){
						in_header = false;
						if (fdata != NULL){
							int n;
//fprintf (fout,"read debut\n");
//fflush (fout);
							while ((n=inp.read(line,sizeof(line)))>0){
//fprintf (fout,"read %d\n",n);
//fflush (fout);
								char *pt = line;
								char *end = line + n - (len_boundary + 3);
								while (pt < end){
									pt = (char*)memchr (pt,'\n',(end-pt)+1);
//fprintf (fout,"read memchr %p\n",pt);
//fflush (fout);
									if (pt == NULL || pt > end){
										pt = NULL;
										break;
									}
									if (pt[1] == '-'
										&& pt[2] == '-'
										&& memcmp(pt+3,boundary,len_boundary)==0){
										break;
									}else{
										pt++;
									}
								}
								if (pt != NULL && pt < end){
									// Ok, this is the end of the attachement
									int datalen = (pt-line);
									if (datalen > 0) fwrite(line,1,datalen,fdata);
									if (datalen < n) inp.unread (line+datalen,n-datalen);
									break;
								}else{
									fwrite (line,1,n,fdata);
								}
							}
							fclose (fdata);
							fdata = NULL;
							in_header = true;
						}
					}else if (strncasecmp(line,"Content-Disposition:",20)==0){
						SSTRING filename;
						http_pickname (line,"name=",variable);
						http_pickname (line,"filename=",filename);
						if (filename.is_filled()){
							char templat[30];
							strcpy (templat,"/tmp/tlmpweb-file.XXXXXX");
							int fd = mkstemp (templat);
							if (fd != -1){
								// Remember to delete the file
								tbfile.add (new SSTRING(templat));
								fdata = fdopen (fd,"w");
								// We define a pseudo variable filename:variable
								// we will retrieve later. It contains
								// the name of the temp file
								SSTRING *s = new SSTRING;
								s->setfromf ("filename:%s",variable.get());
								tbvar.add (s);
								tbval.add (new SSTRING(templat));
								tbvar.add (new SSTRING(variable));
								tbval.add (new SSTRING(filename));
							}else{
								// Can't save the file, forget the variable
								variable.setfrom ("");
							}
						}
					}else if (strncasecmp(line,"Content-type:",13)==0){
					}
				}else if (fdata != NULL){
					fputs (line,fdata);
				}else{
					int len = strlen(line);
					if (len>0 && line[len-1] == '\n') line[--len] = '\0';
					if (len>0 && line[len-1] == '\r') line[len-1] = '\0';
					if (accum.is_filled()){
						accum.appendf ("\n%s",line);
					}else{
						accum.append (line);
					}
				}
			}
			if (fdata != NULL) fclose (fdata);
		}
			
	}else{
		// Ordinary POST encoding
		char line[10000];
		while (fgets(line,sizeof(line)-1,stdin)!=NULL){
			// fprintf (fout,"POST: %s",line);
			http_parsevar (line);
		}
	}
	// fclose (fout);
}


static int tlmpweb_exec (const char *path)
{
	int ret = -1;
	void *handle = dlopen (path,RTLD_LAZY);
	if (handle == NULL){
		tlmpweb_error ("CGI error: dlopen (%s,RTLD_LAZY) failed (%s)\n"
			,path,dlerror());
	}else{
		{
			char dir[PATH_MAX];
			strncpy (dir,path,sizeof(dir)-1);
			char *pt = strrchr(dir,'/');
			if (pt != NULL){
				*pt = '\0';
				chdir (dir);
			}
		}
		void (*fct)() = (void (*)()) dlsym (handle,"webmain");
		if (fct == NULL){
			tlmpweb_error ("Invalid object: No webmain symbol\n");
		}else{
			http_parsevar_env();
			http_parsevar_post();
			wvar_init();
			(*fct)();
			if (title_done){
				printf ("</body>\n</html>\n");
			}
			ret = 0;
		}
	}
	return ret;
}

void tlmpweb_copy (const char *file)
{
	FILE *fin = fopen (file,"r");
	if (fin == NULL){
		tlmpweb_error ("Can't copy file %s (%s)\n",file,strerror(errno));
	}else{
		char buf[1000];
		while (fgets (buf,sizeof(buf)-1,fin)!=NULL){
			fputs (buf,stdout);
		}
		fclose (fin);
	}
}

void tlmpweb_include (const char *file)
{
	const char *pt = strrchr(file,'.');
	if (pt == NULL){
		tlmpweb_copy (file);
	}else{
		if (strcmp(pt,".hc")==0){
			char cwd[PATH_MAX];
			getcwd(cwd,sizeof(cwd)-1);
			if (file[0] == '/'){
				tlmpweb_exec (file);
			}else{
				char path[PATH_MAX];
				snprintf (path,sizeof(path)-1,"%s/%s",cwd,file);
				tlmpweb_exec (path);
			}
			chdir (cwd);
		}else{
			tlmpweb_copy (file);
		}
	}
}

/*
	Return true if this file name is safe. It does not contain
	any special characters and does not try to jump to another directory
*/
bool tlmpweb_safefname(
	const char *fname,		// File name to validate
	const char *expected_ext)	// Expected file name extension or NULL
								// if it does not matter.
								// It must contains the dot (.jpeg)
{
	bool ret = false;
	if (fname[0] != '\0'
		&& strcmp(fname,"..")!=0
		&& strcmp(fname,".")!=0){
		ret = true;
		const char *pt = fname;
		while (*pt != '\0'){
			if (!isalpha(*pt)
				&& !isdigit(*pt)
				&& *pt != '_'
				&& *pt != '.'){
				ret = false;
				break;
			}
			pt++;
		}
		if (ret && expected_ext != NULL){
			int len_fname = strlen(fname);
			int len_ext   = strlen(expected_ext);
			if (len_fname <= len_ext){
				ret = false;
			}else if (strcmp(fname+len_fname-len_ext,expected_ext)!=0){
				ret = false;
			}
		}				
	}
	return ret;
}
		

/*
	Return the value of a cookie sent by the browser
*/
const char *tlmpweb_getcookie(const char *name)
{
	static char retstr[200];
	const char *ret = NULL;
	const char *env = getenv ("HTTP_COOKIE");
	if (env != NULL){
		int len = strlen(name);
		while (*env != '\0'){
			if (strncmp(env,name,len)==0 && env[len] == '='){
				// Got it
				env += len+1;
				const char *end = strchr(env,';');
				if (end == NULL){
					http_decode (env,retstr,sizeof(retstr));
				}else{
					char tmp[strlen(env)+1];
					strcpy (tmp,env);
					char *pt = strchr(tmp,';');
					*pt = '\0';
					http_decode (tmp,retstr,sizeof(retstr));
				}
				ret = retstr;
				break;
			}else{
				env = strchr(env,';');
				if (env == NULL) break;
				env = str_skip(env+1);
			}
		}
	}
	return ret;
}

/*
	Return true if we are in HTTPS mode
*/
bool tlmpweb_is_secure()
{
	const char *pt = getenv ("HTTPS");
	return pt != NULL && stricmp(pt,"on")==0;
}

/*
	Force the client to come back in https mode
*/
bool tlmpweb_force_secure()
{
	bool ret = tlmpweb_is_secure();
	if (!ret){
		printf ("Location: https://%s%s/%s\n",tlmpweb_curhost()
			,tlmpweb_curdir()
			,tlmpweb_curpage());
		tlmpweb_doctype ("text/html");
	}
	return ret;
}

int main (int argc, char *argv[])
{
	int ret = -1;
	linuxconf_loadlibmsg("/usr/lib/linuxconf-lib/help");
	translat_load ("/usr/lib/tlmp/help","tlmpweb", "eng");
	tlmp_seterrorfct (tlmpweb_error0);
	#if 0
		FILE *fout = fopen ("/tmp/toto.log","a");
		if (fout != NULL){
			char cwd[1000];
			getcwd (cwd,sizeof(cwd)-1);
			fprintf (fout,"cwd=%s\n",cwd);
			for (int i=0; i<argc; i++){
				fprintf (fout,"argv[%d]=:%s:\n",i,argv[i]);
			}
			extern char **environ;
			for (int i=0; environ[i] != NULL; i++){
				fprintf (fout,"environ[%d]=:%s:\n",i,environ[i]);
			}
			fclose (fout);
		}
	#endif
	setvbuf (stdout,NULL,_IOFBF,8192);
	const char *request = getenv ("REQUEST_URI");
	if (request != NULL && strncmp(request,"/cgi-bin/tlmpweb/",17)==0){
		// Accessing an .hc page in cgi-bin
		char cwd[1000];
		getcwd (cwd,sizeof(cwd)-1);
		curdir = cwd;
		char page[PATH_MAX];
		snprintf (page,sizeof(page)-1,"%s/%s",cwd,request+17);
		char *pt = strchr (page,'?');
		if (pt != NULL) *pt = '\0';
		tlmpweb_loadlibs ();
		ret = tlmpweb_exec (page);
	}else{
		const char *path = getenv ("PATH_TRANSLATED");
		if (path == NULL){
			tlmpweb_error ("CGI error: No PATH_TRANSLATED variable\n");
		}else{
			char dir[PATH_MAX];
			tlmpweb_setdir (path,dir);
			curdir = dir;
			tlmpweb_loadlibs ();
			ret = tlmpweb_exec (path);
		}
	}
	for (int i=0; i<tbfile.getnb(); i++) unlink (tbfile.getitem(i)->get());
	fflush (stdout);
	if (!header_done){
		tlmpweb_doctype ("text/html");
	}
	return ret;
}

